# Third party
from litellm import embedding


class Model(object):
    def __init__(self, model: str, batch_size: int = 100, **kwargs):
        self.model = model
        self.batch_size = batch_size
        self.kwargs = kwargs
        self.adapter = None

    def set_adapter(self, adapter):
        self.adapter = adapter

    def get_embeddings(self, inputs: list[str]) -> list[list[float]]:
        embeddings = []
        for i in range(0, len(inputs), self.batch_size):
            response = embedding(
                self.model, inputs[i : i + self.batch_size], **self.kwargs
            )

            for result in sorted(response["data"], key=lambda d: d["index"]):
                vector = result["embedding"]
                if self.adapter:
                    vector = vector @ self.adapter

                embeddings.append(vector)

        return embeddings

    def get_embedding(self, input: str) -> list[float]:
        return self.get_embeddings([input])[0]



# Third party
import numpy as np


def cosine_similarity(embedding1: list, embedding2: list) -> float:
    vec1 = np.array(embedding1)
    vec2 = np.array(embedding2)

    dot_product = np.dot(vec1, vec2)
    norm1 = np.linalg.norm(vec1)
    norm2 = np.linalg.norm(vec2)

    if norm1 == 0 or norm2 == 0:
        return 0.0

    return dot_product / (norm1 * norm2)


def accuracy_and_stderr(
    cosine_similarity: float, labeled_similarity: int
) -> tuple[float]:
    # calculate accuracy (and its standard error) of predicting label=1 if similarity>x
    # x is optimized by sweeping from -1 to 1 in steps of 0.01

    accuracies = []
    for threshold_thousandths in range(-1000, 1000, 1):
        threshold = threshold_thousandths / 1000
        total = 0
        correct = 0
        for cs, ls in zip(cosine_similarity, labeled_similarity):
            total += 1
            if cs > threshold:
                prediction = 1
            else:
                prediction = -1
            if prediction == ls:
                correct += 1
        accuracy = correct / total
        accuracies.append(accuracy)
    a = max(accuracies)
    n = len(cosine_similarity)
    standard_error = (a * (1 - a) / n) ** 0.5  # standard error of binomial
    return a, standard_error



# Standard library
import asyncio

# Third party
import json_repair
import pandas as pd
from tqdm.asyncio import tqdm_asyncio
from litellm import completion, acompletion
from sklearn.model_selection import train_test_split

# Local
try:
    from weightgain.Model import Model
    from weightgain.utilities import cosine_similarity
except ImportError:
    from .Model import Model
    from utilities import cosine_similarity


#########
# HELPERS
#########


TEST_FRACTION = 0.8
RANDOM_SEED = 123
NEGATIVES_PER_POSITIVE = 1

CHUNKS_PROMPT = """Your task is to generate synthetic data for fine-tuning an embedding model. \
This data will be used to improve retrieval performance in a question-answering system. \
Your goal is to create diverse, realistic text chunks based on a given prompt.

Here is the prompt you should base your chunks on:

<chunk_prompt>
{prompt}
</chunk_prompt>

You need to generate the following number of chunks:

<chunk_count>
{n}
</chunk_count>

Follow these instructions when generating your chunks:

<instructions>
- Read and understand the chunk prompt carefully.
- Generate <chunk_count> unique text chunks based on the chunk prompt.
- Ensure each chunk is diverse and represents a unique aspect or feature of the prompt.
- Each chunk should be realistic and based on real-world situations related to the prompt.
- Avoid repetition or highly similar chunks.
</instructions>"""

QA_PROMPT = """Your task is to generate synthetic questions based on the provided context. \
These questions will be used to fine-tune an embedding model to improve retrieval against this context.

Here is the context you should base your questions on:

<context>
{context}
</context>

You need to generate the following number of questions:

<question_count>
{n}
</question_count>

Follow these instructions when generating your questions:

<instructions>
- Read and analyze the provided context carefully.
- Generate <question_count> unique questions based on the context.
- Base your questions solely on the information in the context. DO NOT use prior knowledge.
- Ensure that your questions are diverse and cover different aspects of the context.
</instructions>"""


def generate_chunks(prompt: str, model: str, n: int = 100) -> list[str]:
    print(f"Generating chunks")

    response = completion(
        model=model,
        messages=[
            {"role": "user", "content": CHUNKS_PROMPT.format(prompt=prompt, n=n)}
        ],
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "chunks_response",
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {
                        "chunks": {
                            "type": "array",
                            "items": {
                                "type": "string",
                                "description": "A text chunk based on the <chunk_prompt>.",
                            },
                            "description": f"List of text chunks based on the <chunk_prompt>. len(chunks) >= {n}.",
                        }
                    },
                    "required": ["chunks"],
                    "additionalProperties": False,
                },
            },
        },
    )
    response = response.choices[0].message.content
    response = json_repair.loads(response)
    chunks = response["chunks"][:n]

    if len(chunks) < n:
        # TODO: Generate more until we reach n
        print(f"Warning: Generated only {len(chunks)} chunks. Expected {n}.")

    return chunks


async def generate_dataset(
    chunks: list[str], model: str, n_per_chunk: int = 100
) -> pd.DataFrame:
    tasks = []
    for chunk in chunks:
        task = acompletion(
            model=model,
            messages=[
                {
                    "role": "user",
                    "content": QA_PROMPT.format(context=chunk, n=n_per_chunk),
                }
            ],
            response_format={
                "type": "json_schema",
                "json_schema": {
                    "name": "questions_response",
                    "strict": True,
                    "schema": {
                        "type": "object",
                        "properties": {
                            "questions": {
                                "type": "array",
                                "items": {
                                    "type": "string",
                                    "description": "A question based on the <context>.",
                                },
                                "description": f"List of questions based on the <context>. len(questions) >= {n_per_chunk}.",
                            },
                        },
                        "required": ["questions"],
                        "additionalProperties": False,
                    },
                },
            },
        )
        tasks.append(task)

    results = await tqdm_asyncio.gather(
        *tasks,
        desc=f"Generating questions",
    )

    dataset = []
    for chunk_id, (chunk, result) in enumerate(zip(chunks, results)):
        result = json_repair.loads(result.choices[0].message.content)
        questions = result["questions"][:n_per_chunk]

        if len(questions) < n_per_chunk:
            print(
                f"Warning: Chunk {chunk_id} generated only {len(questions)} questions."
            )

        dataset += [(chunk_id, question, chunk, 1) for question in questions]

    df = pd.DataFrame(dataset, columns=["chunk_id", "text_1", "text_2", "label"])
    return df


def split_dataset(dataset: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
    if "chunk_id" in dataset.columns:
        chunk_ids = dataset["chunk_id"].unique()
        train_ids, test_ids = train_test_split(
            chunk_ids, test_size=TEST_FRACTION, random_state=RANDOM_SEED
        )

        train_df = dataset[dataset["chunk_id"].isin(train_ids)].copy()
        test_df = dataset[dataset["chunk_id"].isin(test_ids)].copy()
    else:
        train_df, test_df = train_test_split(
            dataset,
            test_size=TEST_FRACTION,
            random_state=RANDOM_SEED,
        )

    train_df.loc[:, "dataset"] = "train"
    test_df.loc[:, "dataset"] = "test"

    return train_df, test_df


def add_negatives(dataset: pd.DataFrame) -> pd.DataFrame:
    texts = set(dataset["text_1"].values) | set(dataset["text_2"].values)
    all_pairs = {(t1, t2) for t1 in texts for t2 in texts if t1 < t2}
    positive_pairs = set(
        tuple(text_pair) for text_pair in dataset[["text_1", "text_2"]].values
    )
    negative_pairs = all_pairs - positive_pairs
    negatives_dataset = pd.DataFrame(list(negative_pairs), columns=["text_1", "text_2"])
    negatives_dataset["label"] = -1
    negatives_dataset["dataset"] = dataset["dataset"].iloc[0]

    dataset = pd.concat(
        [
            dataset,
            negatives_dataset.sample(
                n=len(dataset) * NEGATIVES_PER_POSITIVE, random_state=RANDOM_SEED
            ),
        ]
    )
    return dataset


def add_embeddings(dataset: pd.DataFrame, model: Model):
    print("Calculating embeddings")

    for column in ["text_1", "text_2"]:
        dataset[f"{column}_embedding"] = model.get_embeddings(dataset[column])

    dataset["cosine_similarity"] = dataset.apply(
        lambda row: cosine_similarity(row["text_1_embedding"], row["text_2_embedding"]),
        axis=1,
    )


def add_negatives_and_embeddings(dataset: pd.DataFrame, model: Model) -> pd.DataFrame:
    train_df, test_df = split_dataset(dataset)
    train_df = add_negatives(train_df)
    test_df = add_negatives(test_df)
    dataset = pd.concat([train_df, test_df])
    add_embeddings(dataset, model)
    return dataset


######
# MAIN
######


class Dataset(object):
    def __init__(self, df: pd.DataFrame):
        self.df = df

    def __getitem__(self, key):
        return self.df[key]

    def __setitem__(self, key, value):
        self.df[key] = value

    def apply(self, func, axis=0, **kwargs):
        return self.df.apply(func, axis=axis, **kwargs)

    @classmethod
    def from_chunks(
        cls, chunks: list[str], model: Model, llm: str, n_queries_per_chunk: int = 1
    ) -> "Dataset":
        df = asyncio.run(generate_dataset(chunks, llm, n_queries_per_chunk))
        df = add_negatives_and_embeddings(df, model)
        return cls(df)

    @classmethod
    def from_synthetic_chunks(
        cls,
        prompt: str,
        model: Model,
        llm: str,
        n_chunks: int = 100,
        n_queries_per_chunk: int = 1,
    ) -> "Dataset":
        chunks = generate_chunks(prompt, llm, n_chunks)

        try:
            loop = asyncio.get_running_loop()
            df = loop.run_until_complete(
                generate_dataset(chunks, llm, n_queries_per_chunk)
            )
        except RuntimeError:
            df = asyncio.run(generate_dataset(chunks, llm, n_queries_per_chunk))

        df = add_negatives_and_embeddings(df, model)
        return cls(df)

    @classmethod
    def from_pairs(cls, text_pairs: list[tuple[str, str]], model: Model) -> "Dataset":
        text_pairs = [
            (index, text_1, text_2, 1)
            for index, (text_1, text_2) in enumerate(text_pairs)
        ]
        df = pd.DataFrame(text_pairs, columns=["id", "text_1", "text_2", "label"])
        df = add_negatives_and_embeddings(df, model)
        return cls(df)



# Standard library
import random
from typing import Optional

# Third party
import torch
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

# Local
try:
    from weightgain.Dataset import Dataset
    from weightgain.utilities import cosine_similarity, accuracy_and_stderr
except ImportError:
    from .Dataset import Dataset
    from utilities import cosine_similarity, accuracy_and_stderr


#########
# HELPERS
#########


def tensors_from_df(df: pd.DataFrame) -> tuple[torch.tensor]:
    embedding1 = np.stack(np.array(df["text_1_embedding"].values))
    embedding2 = np.stack(np.array(df["text_2_embedding"].values))
    similarity = np.stack(np.array(df["label"].astype("float").values))

    embedding1 = torch.from_numpy(embedding1).float()
    embedding2 = torch.from_numpy(embedding2).float()
    similarity = torch.from_numpy(similarity).float()

    return embedding1, embedding2, similarity


def model(
    embedding1: torch.tensor,
    embedding2: torch.tensor,
    matrix: torch.tensor,
    dropout: float,
) -> float:
    embedding1 = torch.nn.functional.dropout(embedding1, p=dropout)
    embedding2 = torch.nn.functional.dropout(embedding2, p=dropout)
    embedding1 = embedding1 @ matrix
    embedding2 = embedding2 @ matrix
    similarity = torch.nn.functional.cosine_similarity(embedding1, embedding2)
    return similarity


def mse_loss(predictions: torch.tensor, targets: torch.tensor) -> torch.tensor:
    difference = predictions - targets
    return torch.sum(difference * difference) / difference.numel()


def biased_embedding(embedding: list[float], matrix: torch.tensor) -> np.array:
    embedding = torch.tensor(embedding).float()
    embedding = embedding @ matrix
    embedding = embedding.detach().numpy()
    return embedding


def apply_matrix_to_df(matrix: torch.tensor, df: pd.DataFrame):
    for column in ["text_1_embedding", "text_2_embedding"]:
        df[f"{column}_custom"] = df[column].apply(lambda x: biased_embedding(x, matrix))

    df["cosine_similarity_custom"] = df.apply(
        lambda row: cosine_similarity(
            row["text_1_embedding_custom"], row["text_2_embedding_custom"]
        ),
        axis=1,
    )


######
# MAIN
######


class Adapter(object):
    def __init__(self, matrix: np.ndarray, results: Optional[pd.DataFrame] = None):
        self.matrix = matrix
        self.results = results

    def __matmul__(self, embedding) -> np.ndarray:
        if isinstance(embedding, list):
            embedding = np.array(embedding)

        return embedding @ self.matrix

    def to_csv(self, path: str):
        if self.results:
            raise Exception("Cannot save Adapter without training results")

        self.results.to_csv(path, index=False)

    def plot_sim_dists(self):
        fig1 = px.histogram(
            self.results,
            x="cosine_similarity",
            color="label",
            barmode="overlay",
            width=500,
            facet_row="dataset",
        )
        fig2 = px.histogram(
            self.results,
            x="cosine_similarity_custom",
            color="label",
            barmode="overlay",
            width=500,
            facet_row="dataset",
        )
        fig1.show()
        fig2.show()

    def show_report(self, save_path: str = None):
        if self.results is None:
            raise Exception("Cannot generate report without training results")

        # Create a 2x2 subplot figure
        fig = make_subplots(
            rows=2,
            cols=2,
            subplot_titles=(
                "Training and Test Loss",
                "Training and Test Accuracy",
                "Original Cosine Similarity Distribution",
                "Adapted Cosine Similarity Distribution",
            ),
        )

        # Add loss plot
        for dataset_type in self.results["type"].unique():
            data = self.results[self.results["type"] == dataset_type]
            fig.add_trace(
                go.Scatter(
                    x=data["epoch"],
                    y=data["loss"],
                    mode="lines",
                    name=f"{dataset_type} loss",
                ),
                row=1,
                col=1,
            )

        # Add accuracy plot
        for dataset_type in self.results["type"].unique():
            data = self.results[self.results["type"] == dataset_type]
            fig.add_trace(
                go.Scatter(
                    x=data["epoch"],
                    y=data["accuracy"],
                    mode="lines",
                    name=f"{dataset_type} accuracy",
                ),
                row=1,
                col=2,
            )

        # Add original similarity distribution
        for label_value in [-1, 1]:
            label_subset = self.dataset[self.dataset["label"] == label_value]
            if not label_subset.empty:
                fig.add_trace(
                    go.Histogram(
                        x=label_subset["cosine_similarity"],
                        name=f"Original {dataset_type} (label={label_value})",
                        opacity=0.7,
                        nbinsx=30,
                        marker_color="blue" if label_value == 1 else "red",
                    ),
                    row=2,
                    col=1,
                )

        # Add custom similarity distribution
        for label_value in [-1, 1]:
            label_subset = self.dataset[self.dataset["label"] == label_value]
            if not label_subset.empty:
                fig.add_trace(
                    go.Histogram(
                        x=label_subset["cosine_similarity_custom"],
                        name=f"Custom {dataset_type} (label={label_value})",
                        opacity=0.7,
                        nbinsx=30,
                        marker_color="blue" if label_value == 1 else "red",
                    ),
                    row=2,
                    col=2,
                )

        # Update layout
        fig.update_layout(
            height=800, width=1000, title_text="Training Report", barmode="overlay"
        )

        if save_path:
            fig.write_html(save_path)
            print(f"Report saved to {save_path}")
        else:
            fig.show()

    @classmethod
    def from_file(cls, path: str) -> "Adapter":
        matrix = np.load(path)
        return cls(matrix)

    @classmethod
    def train(
        cls,
        dataset: Dataset,
        batch_size: int = 100,
        max_epochs: int = 100,
        learning_rate: float = 100.0,
        dropout: float = 0.0,
    ) -> "Adapter":
        embedding_len = len(dataset["text_1_embedding"].values[0])

        emb1_train, emb2_train, sim_train = tensors_from_df(
            dataset[dataset["dataset"] == "train"]
        )
        emb1_test, emb2_test, sim_test = tensors_from_df(
            dataset[dataset["dataset"] == "test"]
        )

        train_loader = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(emb1_train, emb2_train, sim_train),
            batch_size=batch_size,
            shuffle=True,
        )
        matrix = torch.randn(embedding_len, embedding_len, requires_grad=True)

        best_acc, best_matrix = 0, matrix
        epochs, types, losses, accuracies = [], [], [], []
        for epoch in range(1, 1 + max_epochs):
            for emb1, emb2, target_sim in train_loader:
                # Generate prediction and calculate loss
                pred_sim = model(emb1, emb2, matrix, dropout)
                loss = mse_loss(pred_sim, target_sim)
                loss.backward()  # Backpropagate loss

                # Update matrix
                with torch.no_grad():
                    matrix -= matrix.grad * learning_rate
                    matrix.grad.zero_()

            # Calculate test loss
            test_preds = model(emb1_test, emb2_test, matrix, dropout)
            test_loss = mse_loss(test_preds, sim_test)

            # Compute new embeddings + similarities
            apply_matrix_to_df(matrix, dataset)

            # Calculate test accuracy
            for dataset_type in ["train", "test"]:
                data = dataset[dataset["dataset"] == dataset_type]
                accuracy, stderr = accuracy_and_stderr(
                    data["cosine_similarity_custom"], data["label"]
                )

                epochs.append(epoch)
                types.append(dataset_type)
                losses.append(
                    loss.item() if dataset_type == "train" else test_loss.item()
                )
                accuracies.append(accuracy)

                if accuracy > best_acc and dataset_type == "test":
                    best_acc = accuracy
                    best_matrix = matrix

                print(
                    f"Epoch {epoch}/{max_epochs}: {dataset_type} accuracy: {accuracy:0.1%} ± {1.96 * stderr:0.1%}"
                )

        results = pd.DataFrame(
            {
                "epoch": epochs,
                "type": types,
                "loss": losses,
                "accuracy": accuracies,
            }
        )
        results["batch_size"] = batch_size
        results["max_epochs"] = max_epochs
        results["learning_rate"] = learning_rate
        results["dropout"] = dropout

        adapter = cls(best_matrix.detach().numpy(), results)
        adapter.dataset = dataset
        return adapter



try:
    from weightgain.Adapter import Adapter
    from weightgain.Dataset import Dataset
    from weightgain.Model import Model
except ImportError:
    from .Adapter import Adapter
    from .Dataset import Dataset
    from .Model import Model
